/* $Id: stream.c,v 1.16 1997/08/06 06:11:08 phil Exp $ */

#include "h.h"
#include "snotypes.h"
#include "libret.h"
#include "syntab.h"

#include "macros.h"
#include "equ.h"
#include "res.h"
#include "data.h"			/* for STYPE */

#ifdef DEBUG
#define DEBUGF(lev, x) if (lev <= DEBUG) printf x
#else  /* DEBUG not defined */
#define DEBUGF(lev, x)
#endif /* DEBUG not defined */


#define ACCEPT() { cp++; len--; }

enum stream_ret
stream( sp1, sp2, tp )
    struct spec *sp1;			/* OUT: prefix */
    struct spec *sp2;			/* IN: string OUT: remainder */
    struct syntab *tp;
{
    unsigned char *cp;
    int len;
    enum stream_ret ret;

    len = S_L(sp2);
    cp = (unsigned char *)S_SP(sp2);
    D_A(STYPE) = 0;			/* XXX in case no puts?? */

    DEBUGF(1,("stream"));
    DEBUGF(10,(" '%*s'", len, cp));
    DEBUGF(1,(" table %s\n", tp->name ));

    while (len > 0) {
	const struct acts *ap;
	int_t tok;

	ap = tp->actions + tp->chrs[*cp];

	DEBUGF(2,(" '%c' (%d)", *cp, *cp ));

	/* token can never occur with CONTIN or ERROR? */
	tok = ap->put;
	if (tok) {
	    DEBUGF(2,(" put %d", tok ));
	    D_A(STYPE) = tok;
	}

#ifdef DEBUG
	switch (ap->act) {
	case AC_CONTIN:
	    DEBUGF(2,(" CONTIN\n"));
	    break;
	case AC_STOP:
	    DEBUGF(2,(" STOP\n"));
	    break;
	case AC_STOPSH:
	    DEBUGF(2,(" STOPSH\n"));
	    break;
	case AC_ERROR:
	    DEBUGF(2,(" ERROR\n"));
	    break;
	case AC_GOTO:
	    DEBUGF(2,(" goto %s\n", ap->go->name));
	    break;
	}
#endif /* DEBUG defined */

	switch (ap->act) {
	case AC_CONTIN:
	    break;
	case AC_STOP:
	    cp++; len--;		/* accept */
	    /* FALL */
	case AC_STOPSH:
	    ret = ST_STOP;
	    goto break_loop;
	case AC_ERROR:
	    D_A(STYPE) = 0;
	    return ST_ERROR;		/* immediate return! */
	case AC_GOTO:
	    tp = ap->go;
	    break;
	}
	cp++; len--;			/* accept */
    }
    /* here when out of subject */
    ret = ST_EOS;

 break_loop:
    len = S_L(sp2) - len;		/* get match length */

    _SPEC(sp1) = _SPEC(sp2);		/* copy spec for prefix */
    S_L(sp1) = len;			/* set prefix length */

    if (ret != ST_EOS)
	S_O(sp2) += len;		/* bump suffix offset */

    S_L(sp2) -= len;			/* adjust suffix length */

    return ret;
}

/* 10/28/93 */
void
clertb(tp, act, sp)
    struct syntab *tp;
    enum action act;
{
    int i, j;
    char *cp;
    union {
	long l;
	char c[sizeof(long)];
    } u;
    register long *lp, l;

    /* find action index in list (SNABTB has one of each action type) */
    for (j = 0; ; j++)
	if (tp->actions[j].act == act)
	    break;

    /* setup long's worth of chars */
    for (i = 0; i < sizeof(u.c); i++)
	u.c[i] = j;

    /* stamp out long's with unrolled loop */
    l = u.l;
    lp = (long *) tp->chrs;
    i = CHARSET / sizeof(u.c) / 8;
    do {
	/* most ISA's have index+offset; RISC's tend not to have autoinc */
	lp[0] = l;
	lp[1] = l;
	lp[2] = l;
	lp[3] = l;
	lp[4] = l;
	lp[5] = l;
	lp[6] = l;
	lp[7] = l;
	lp += 8;
    } while (--i != 0);
}

/* 10/28/93 */
void
plugtb(tp, act, sp)
    struct syntab *tp;
    enum action act;
    struct spec *sp;
{
    unsigned char *cp;
    int len;
    int j;

    len = S_L(sp);
    cp = (unsigned char *)S_SP(sp);

    /* find action index in list (SNABTB has one of each action type) */
    for (j = 0; ; j++)
	if (tp->actions[j].act == act)
	    break;

    while (len-- > 0) {
	tp->chrs[*cp++] = j;
    }
}

/* 8/5/97 [PLB59] */
int
any(sp, dp)
    struct spec *sp;			/* subject */
    struct descr *dp;			/* (not)any arg str */
{
    register unsigned char c, *cp;
    struct descr *vp;
    register int i;

    c = *(unsigned char *)S_SP(sp);	/* get next subject char */
    vp = (struct descr *)D_A(dp);	/* ptr to char set var */
    cp = ((unsigned char *)vp) + BCDFLD; /* ptr to char set */
    i = D_V(vp);			/* length of char set */

    while (i > 0) {
	if (*cp == c)
	    return 1;
	cp++;
	i--;
    }
    return 0;
}
